/* LICENSE=GPL */

/* iFeel force feedback uinput driver
** written by emard@softhome.net
*/

/* compile with libusb, use gcc -lusb -lpthread */

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <pthread.h>
#include <signal.h>
#include <asm/types.h>
#include <linux/input.h>
#include <linux/uinput.h>
#include <usb.h>

pthread_mutex_t mutex1 = PTHREAD_MUTEX_INITIALIZER;

#include "usbid.h"

#define TIMEOUT 3000
#define WAITINBETWEEN 500000

/* From HID specification */
#define SET_REQUEST 0x21
#define SET_REPORT 0x09 
#define OUTPUT_REPORT_TYPE 0x02

/* how many controlling threads */
#define THREADS 2

int rc[THREADS];
pthread_t thread[THREADS];

/* Global list of FDs each object is interested in,
** and FDs that there have been activity on. Each object
** adds itself to the fd_request_* lists on creation and removes
** itself on deletion, and checks the fd_* lists on poll.
*/
static fd_set fd_request_read, fd_request_write, fd_request_except;
static fd_set fd_read, fd_write, fd_except;
int fd_count = 0;

struct ifeel {
  struct usb_device *dev; /* libusb dev */
  struct usb_dev_handle *mouse;
  int uifd; /* os-side uinput fd (read/write) */
  struct uinput_user_dev device;
  unsigned int strength, delay, count;
  int countdown; /* effect countdown counter */
  struct timeval timer_cont; /* continuation timer */
  unsigned char led; /* virtual LED */
  char data[4]; /* previous data received from the mouse */
};

struct ifeel ifeel;


/* This initializes libusb and tracks down the device */
static struct usb_device *init_device()
{
  struct usb_bus *bus;
  struct usb_device *dev;
  
  usb_init();
  usb_find_busses();
  usb_find_devices();
  
  for (bus = usb_busses; bus; bus = bus->next)
  {
    for (dev = bus->devices; dev; dev = dev->next)
    {
      if ((dev->descriptor.idVendor == USB_MOUSE_VENDOR) &&
          (dev->descriptor.idProduct == USB_MOUSE_PRODUCT))
        return dev;
    }
  }
  return NULL;
}


/* register the mouse in uinput system as
** pointing device with
** a number of keys, relative pointers and leds
** then it will automatically appear as /dev/input/mouse*
*/
int ifeel_register(struct ifeel *mouse)
{
  int aux, fd;

  /* libusb initialization */

  mouse->dev = init_device();
  if(!mouse->dev)
  {
    printf("Unable to find device.\n");
    return 1;
  }
  
  /* Open device, note: we want interface zero so we only have to open
     once */
  mouse->mouse = usb_open(mouse->dev);
  if(!mouse->mouse)
  {
    printf("Unable to open device.\n");
    return 1;
  }
  
  if(usb_set_configuration(mouse->mouse, mouse->dev->config->bConfigurationValue))
  {
    printf("Failed to set config 1: EBUS: %d ENONM: %d\n", EBUSY, ENOMEM);
    return 1;
  }

  /* The Mouse is interface zero */		
  if(usb_claim_interface(mouse->mouse, 0) < 0)
  {
    printf("Unable to claim interface 0.\n");
    return 1;
  }

  /* uinput initialization */
  FD_ZERO(&fd_request_read);
  FD_ZERO(&fd_request_write);
  FD_ZERO(&fd_request_except);
  
  /* open uinput device file */
  fd = open("/dev/input/uinput", O_RDWR);
  if (fd < 0) {
    perror("open");
    return -1;
  }

  mouse->uifd = fd;

  /* sets the name of our device */
  strcpy(mouse->device.name, "Logitech Inc. iFeel Mouse");

  /* sets maximum number of simultaneous force feedback effects */
  mouse->device.ff_effects_max = 1;
  
  /* its bus */
  mouse->device.id.bustype = BUS_USB;
  
  /* and id/vendor id/product id/version
  ** we will create with uinput
  */
  mouse->device.id.vendor = 0x46d;
  mouse->device.id.product = 0xc030;
  mouse->device.id.version = 257;
  
  /* inform that we'll generate key events */
  ioctl(fd, UI_SET_EVBIT, EV_KEY);
  
  /* inform that we'll generate relative mouse pointer events */
  ioctl(fd, UI_SET_EVBIT, EV_REL);
  
  /* inform that we have force feedback */
  ioctl(fd, UI_SET_EVBIT, EV_FF);

  /* inform that we have LED */
  ioctl(fd, UI_SET_EVBIT, EV_LED);
  
  /* set keyboard events we can generate (in this case, all) */
  for (aux = 1; aux < 207; aux++)
    ioctl(fd, UI_SET_KEYBIT, aux);
  
  /* set mouse events we can generate */
  ioctl(fd, UI_SET_RELBIT, REL_X);
  ioctl(fd, UI_SET_RELBIT, REL_Y);
  ioctl(fd, UI_SET_RELBIT, REL_WHEEL);
  ioctl(fd, UI_SET_KEYBIT, BTN_LEFT);
  ioctl(fd, UI_SET_KEYBIT, BTN_RIGHT);
  ioctl(fd, UI_SET_KEYBIT, BTN_MIDDLE);

  /* set force feedback events we can play */
  ioctl(fd, UI_SET_FFBIT, FF_SINE);
  ioctl(fd, UI_SET_FFBIT, FF_GAIN);

  /* set LED we can display */
  ioctl(fd, UI_SET_LEDBIT, LED_MISC);
  /* ioctl(fd, UI_SET_LEDBIT, LED_SLEEP); */
  
  /* write down information for creating a new device */
  if (write(fd, &(mouse->device), sizeof(struct uinput_user_dev)) < 0) {
    perror("write");
    close(fd);
    return 1;
  }

  /* actually creates the device */
  ioctl(fd, UI_DEV_CREATE);
  
  /* now we can register our uinput device for select() */
  FD_SET(fd, &fd_request_read);
  if(fd >= fd_count)
    fd_count = fd + 1;

  /* start with no effect */
  mouse->countdown = 0;

  return 0;
}


/* wait for file descriptors.
** waits for vibration commands
** from uinput system, or for timeouts
** in order to continue lengthy vibrations
*/
int ifeel_wait(struct ifeel *mouse)
{
  int n;
  struct timeval *timeout;

  timeout = &(mouse->timer_cont);

  if(timeout->tv_sec == 0 && timeout->tv_usec == 0)
    timeout = NULL;

  memcpy(&fd_read, &fd_request_read, sizeof(fd_set));
  memcpy(&fd_write, &fd_request_write, sizeof(fd_set));
  memcpy(&fd_except, &fd_request_except, sizeof(fd_set));
  n = select(fd_count, &fd_read, &fd_request_write, &fd_request_except, timeout );

  return 0;
}


/* unregister the mouse from uinput system
*/
int ifeel_unregister(struct ifeel *mouse)
{
  /* uinput de-initialization */
  close(mouse->uifd);

  /* libusb de-initialization */
  usb_close(mouse->mouse);

  return 0;
}


int ifeel_vibrate(struct ifeel *mouse, int play)
{
  unsigned char vibra_cmd[] = { 0x11, 0x0a,   90, 5, 0x00, 100, 0x00 };
  int delay;

  /* create the usb command for vibration */
  /* safety measure minimum delay=4 to avoid frequencies above 250Hz */
  vibra_cmd[2] = mouse->strength > 255 ? 255 : mouse->strength;
  vibra_cmd[3] = mouse->delay < 4 ? 4 : mouse->delay;
  if(mouse->countdown)
  {
    if(mouse->countdown > 255)
    {
      vibra_cmd[5] = 255;
      mouse->countdown -= 255;
    }
    else
    {
      vibra_cmd[5] = mouse->countdown;
      mouse->countdown = 0;
    }
  }

  mouse->timer_cont.tv_sec = 0;
  mouse->timer_cont.tv_usec = 0;

  if(mouse->countdown && mouse->delay > 0)
  {
    unsigned int microsec = 1000 * vibra_cmd[5] * mouse->delay;
    /* calculate time instance when vibra_cmd will expire,
    ** in order to continue vibration exceeding 255 pulses
    */
    mouse->timer_cont.tv_sec = microsec / 1000000;
    mouse->timer_cont.tv_usec = microsec % 1000000;
  }
  
  /* do we need to stop vibration ? */
  if(mouse->delay == 0 || mouse->count == 0 || mouse->strength == 0 || play <= 0)
  {
    /* stop vibration */
    vibra_cmd[2] = 0;
    vibra_cmd[3] = 20;
    vibra_cmd[5] = 0;

    /* erase any continuation of the effect */
    mouse->countdown = 0;
    mouse->timer_cont.tv_sec = 0;
    mouse->timer_cont.tv_usec = 0;
  }

  /* Send the output reports */
  /* This is a control message as per pg 62 */
  usb_control_msg(mouse->mouse, SET_REQUEST, SET_REPORT,
                  OUTPUT_REPORT_TYPE << 8, 0, (char *)vibra_cmd, sizeof(vibra_cmd), TIMEOUT);

  return 0;
}


/* process uinput event (force feedback effects and LED) */
int uinput_event(struct ifeel *mouse)
{
  struct input_event      event, received_event;
  struct uinput_ff_upload ff_up;
  struct uinput_ff_erase  ff_erase;
  int fd;
  int                     long_count;

  fd = mouse->uifd;

  memset(&ff_up, sizeof(ff_up), 0);

  if(read(fd, &received_event, sizeof(received_event)) == sizeof(received_event))
  {
    switch(received_event.type)
    {
    case EV_FF:
#if 0
      printf("Force feedback request: type=%d code=%d value=%d\n", 
             received_event.type, received_event.code, received_event.value);
#endif
      switch(received_event.code)
      {
      case FF_GAIN:
        printf("FF_GAIN requested\n");
        break;
      default:
        printf("%s effect (delay=%d, strength=%d, count=%d)\n",
               received_event.value > 0 ? "play" : "stop",
               mouse->delay, mouse->strength, mouse->count);
        mouse->countdown = mouse->count; /* this will start effect */
        mouse->timer_cont.tv_sec = 0;
        mouse->timer_cont.tv_usec = 0;
        ifeel_vibrate(mouse, received_event.value);
        break;
      }
      break;
    case EV_LED:
      switch(received_event.code)
      {
      case LED_MISC:
        printf("LED=%d\n", received_event.value);
        mouse->led = received_event.value > 0 ? 1 : 0;
        break;
      }
    case EV_UINPUT:
      switch(received_event.code)
      {
      case UI_FF_UPLOAD:
        printf("Effect upload requested.\n");
        ff_up.request_id = received_event.value;
        if(ioctl(fd, UI_BEGIN_FF_UPLOAD, &ff_up) < 0) {
          perror("ioctl UI_BEGIN_FF_UPLOAD");
          return;
        }
        /* upload effect (ff_up.request_id, ff_up.effect) */
        switch(ff_up.effect.type)
        {
        case FF_PERIODIC:
          switch(ff_up.effect.u.periodic.waveform)
          {
          case FF_SINE:
            mouse->delay = ff_up.effect.u.periodic.period;
            mouse->strength = ff_up.effect.u.periodic.magnitude > 255 ? 255 : ff_up.effect.u.periodic.magnitude;
            mouse->count = (ff_up.effect.u.periodic.envelope.attack_length + ff_up.effect.u.periodic.envelope.fade_length)/mouse->delay;
            printf("Sinewave request_id=%d effect_id=%d period=%dms (%dHz) magnitude=%d length=%dms\n", 
                   ff_up.request_id,
                   ff_up.effect.id,
                   ff_up.effect.u.periodic.period, 1000/ff_up.effect.u.periodic.period,
                   ff_up.effect.u.periodic.magnitude,
                   ff_up.effect.u.periodic.envelope.attack_length + ff_up.effect.u.periodic.envelope.fade_length
              );
            printf("effect uploaded (delay=%d, strength=%d, count=%d)\n",
                   mouse->delay, mouse->strength, mouse->count);
            ff_up.retval = 0;
            ff_up.effect.id = 0;
            break;
          }
          break;
        }
        if(ioctl(fd, UI_END_FF_UPLOAD, &ff_up) < 0) {
          perror("ioctl UI_END_FF_UPLOAD");
          return;
        }
        break;
      case UI_FF_ERASE:
        printf("Effect erase requested.\n");
        ff_erase.request_id = received_event.value;
        if(ioctl(fd, UI_BEGIN_FF_ERASE, &ff_erase) < 0) {
          perror("ioctl UI_BEGIN_FF_ERASE");
        }
        mouse->delay = 0;
        mouse->strength = 0;
        mouse->count = 0;
        printf("request_id=%d effect_id=%d erased.\n", ff_erase.request_id, ff_erase.effect_id);
        ff_erase.retval = 0;
        if(ioctl(fd, UI_END_FF_ERASE, &ff_erase) < 0) {
          perror("ioctl UI_END_FF_UPLOAD");
        }
        break;
      }
    }
  }
  return 0;
}


/* create single thread containing all here */
void *vibra_handler(void *data)
{
  struct ifeel *mouse = (struct ifeel *) data;
  int i, n;

  /* loop processing events */
  for(;;)
  {
    /* wait for events from uinput */
    ifeel_wait(mouse);

    pthread_mutex_lock(&mutex1);
    /* is this uinput event? */
    if(FD_ISSET(mouse->uifd, &fd_read))
      uinput_event(mouse);
    else
    {
      /* no, this could be timer-based continuation */
      if(mouse->countdown)
        ifeel_vibrate(mouse, 1);
    }
    pthread_mutex_unlock(&mutex1);
  }

  return NULL;
}


int key_event(struct ifeel *mouse, char *buffer, int mask, int code)
{
  struct input_event event;

  if( (buffer[0] & mask) != (mouse->data[0] & mask) )
  {
    event.type = EV_KEY;
    event.code = code;
    event.value = (buffer[0] & mask) != 0 ? 1 : 0;
    write(mouse->uifd, &event, sizeof(struct input_event));
    return 1;
  }
  return 0;
}


int axis_event(struct ifeel *mouse, char *buffer, int byte, int code)
{
  struct input_event event;

  if(buffer[byte] != 0)
  {
    event.type = EV_REL;
    event.code = code;
    event.value = buffer[byte];
    write(mouse->uifd, &event, sizeof(struct input_event));
    return 1;
  }
  return 0;
}


void *mouse_handler(void *data)
{
  struct ifeel *mouse = (struct ifeel *) data;
  struct input_event event;
  int i, j, ret, change, fd;
  char buffer[4];

  fd = mouse->uifd;
  for(;;)
  {
    ret = usb_interrupt_read(mouse->mouse, 1, buffer, 4, 500000);
    if(ret < 0) {
      printf("Failure usb_interrupt_read  %d '%s'\n", ret, strerror(-ret));
    }
#if 0
    printf("Read %d bytes: ", ret);
    for(j = 0; j < ret; j++)
      printf("%02x ",
             ((int)buffer[j] & 0x0ff));
    printf("\n");
#endif
    /* decode event */

    /* In vague hallucination I have not heard
    ** master Yoda say:
    **
    ** mutexing this and uinput for
    ** kernel 2.6.13.4 crash prevent could not not
    ** but rather delayed would
    */
    pthread_mutex_lock(&mutex1);
    if(ret == 4)
    {
      change = 0;

      change += key_event(mouse, buffer, 1, BTN_LEFT);
      change += key_event(mouse, buffer, 2, BTN_RIGHT);
      change += key_event(mouse, buffer, 4, BTN_MIDDLE);

      change += axis_event(mouse, buffer, 1, REL_X);
      change += axis_event(mouse, buffer, 2, REL_Y);
      change += axis_event(mouse, buffer, 3, REL_WHEEL);

      /* if there is some change we need to send EV_SYN
      ** this is very important otherwise events will not be
      ** received by the operating system
      */
      if(change)
      {
        event.type = EV_SYN;
        event.code = SYN_REPORT;
        event.value = 0;
        write(fd, &event, sizeof(struct input_event));        
      }

      /* track changes */
      memcpy(mouse->data, buffer, 4);
    }
    pthread_mutex_unlock(&mutex1);
  }

  return NULL;
}


void terminate(int signal)
{
  pthread_cancel(thread[0]);
  pthread_cancel(thread[1]);
}


int main(int argn, char *argv[])
{
  int i, n = THREADS;
  struct ifeel *mouse = &ifeel;

  if(ifeel_register(mouse))
  {
    printf("Cannot open device\n");
    return -1;
  }
  
  printf("User space driver (uinput) for Logitech iFeel mouse.\n");
  /* Create 2 independant threads
  ** one for force feedback handler
  ** second for normal mouse events (kays and movement)
  */

  i = 0;
  if( (rc[i] = pthread_create( &thread[i], NULL, &vibra_handler, (void *) (mouse) )) )
    printf("Thread creation failed: %d\n", rc[i]);

  i = 1;
  if( (rc[i] = pthread_create( &thread[i], NULL, &mouse_handler, (void *) (mouse) )) )
    printf("Thread creation failed: %d\n", rc[i]);

  /* graceful exit handler */
  signal(SIGINT, &terminate);
  signal(SIGTERM, &terminate);

  /* Wait till threads are complete before main continues. Unless we  */
  /* wait we run the risk of executing an exit which will terminate   */
  /* the process and all threads before the threads have completed.   */
  
  pthread_join( thread[0], NULL);
  pthread_join( thread[1], NULL);

  ifeel_unregister(mouse);
  printf("exit.\n");
  
  return 0;
}
